{{- with .Params.functions_and_methods.aliases }}
  {{- $label := "Alias" }}
  {{- if gt (len .) 1 }}
    {{- $label = "Aliases" }}
  {{- end }}
  <p class="font-bold text-dark dark:text-light mt-2">{{ $label }}</p>
  {{- range . }}
    <div class="font-sm font-mono ml-3 sm:ml-6 mt-0 sm:mt-1">
      {{- . -}}
    </div>
  {{- end }}
{{- end -}}
