<meta name="viewport" content="width=device-width, initial-scale=1, viewport-fit=cover">
{{ hugo.Generator }}

{{ if hugo.IsProduction }}
  <meta name="robots" content="index, follow" />
{{ else }}
  <meta name="robots" content="noindex, nofollow" />
{{ end }}

<title>
  {{ with .Title }}{{ . }} |{{ end }}
  {{ .Site.Title }}
</title>

<link rel="preload" href="{{ `fonts/Mulish-VariableFont_wght.ttf` | relURL }}" as="font" type="font/ttf" crossorigin="anonymous" />
<link rel="preload" href="{{ `fonts/Mulish-Italic-VariableFont_wght.ttf` | relURL }}" as="font" type="font/ttf" crossorigin="anonymous" />

<link rel=apple-touch-icon sizes=180x180 href=/apple-touch-icon.png>
<link rel=icon type=image/png href=/favicon-32x32.png sizes=32x32>
<link rel=icon type=image/png href=/favicon-16x16.png sizes=16x16>
<link rel=manifest href=/manifest.json>
<link rel=mask-icon href=/safari-pinned-tab.svg color=#0594cb>

<meta name="turbo-prefetch" content="true">

{{ range .AlternativeOutputFormats -}}
  <link
    rel="{{ .Rel }}"
    type="{{ .MediaType.Type }}"
    href="{{ .RelPermalink | safeURL }}" />
{{ end -}}

<meta name="description" content="{{ or .Description site.Params.description }}">

{{ partial "opengraph/opengraph.html" . }}
{{ partial "schema.html" . }}
{{ partial "twitter_cards.html" . }}

{{ if hugo.IsProduction }}
  {{ partial "helpers/gtag.html" . }}
{{ end }}
