{{ $package := (index .Params 0) }}
{{ $listname := (index .Params 1) }}
{{ $list := (index (index .Site.Data.docs $package) $listname) }}
{{ $fields := after 2 .Params }}


<div class="overflow-x-auto">
  <table>
    <thead>
      <tr>
        {{ range $fields }}
          {{ $s := . }}
          {{ if eq $s "_key" }}
            {{ $s = "type" }}
          {{ end }}
          <th>{{ $s }}</th>
        {{ end }}
      </tr>
    </thead>
    <tbody>
      {{ range $k1, $v1 := $list }}
        <tr>
          {{ range $k2, $v2 := . }}
            {{ $.Scratch.Set $k2 $v2 }}
          {{ end }}
          {{ range $fields }}
            {{ $s := "" }}
            {{ if eq . "_key" }}
              {{ $s = $k1 }}
            {{ else }}
              {{ $s = $.Scratch.Get . }}
            {{ end }}
            <td>{{ $s }}</td>
          {{ end }}
        </tr>
      {{ end }}
    </tbody>
  </table>
</div>
