<meta property="og:title" content="{{ .Title }}">
<meta
  property="og:description"
  content="{{ with .Description }}
    {{ . }}
  {{ else }}
    {{ if .IsPage }}
      {{ .Summary }}
    {{ else }}
      {{ with .Site.Params.description }}{{ . }}{{ end }}
    {{ end }}
  {{ end }}">
<meta
  property="og:type"
  content="{{ if .IsPage }}
    article
  {{ else }}
    website
  {{ end }}">
<meta property="og:url" content="{{ .Permalink }}">

{{- with $.Params.images -}}
  {{- range first 6 . }}
    <meta property="og:image" content="{{ . | absURL }}">
  {{ end -}}
{{- else -}}
  {{- $featured := partial "opengraph/get-featured-image.html" . }}
  {{- with $featured -}}
    <meta property="og:image" content="{{ $featured.Permalink }}">
  {{- else -}}
    {{- with $.Site.Params.images }}
      <meta property="og:image" content="{{ index . 0 | absURL }}">
    {{ end -}}
  {{- end -}}
{{- end -}}

{{- if .IsPage }}
  {{- $iso8601 := "2006-01-02T15:04:05-07:00" -}}
  <meta property="article:section" content="{{ .Section }}">
  {{ with .PublishDate }}
    <meta
      property="article:published_time"
      {{ .Format $iso8601 | printf "content=%q" | safeHTMLAttr }}>
  {{ end }}
  {{ with .Lastmod }}
    <meta
      property="article:modified_time"
      {{ .Format $iso8601 | printf "content=%q" | safeHTMLAttr }}>
  {{ end }}
{{- end -}}

{{- with .Params.audio }}<meta property="og:audio" content="{{ . }}">{{ end }}
{{- with .Params.locale }}
  <meta property="og:locale" content="{{ . }}">
{{ end }}
{{- with .Site.Params.title }}
  <meta property="og:site_name" content="{{ . }}">
{{ end }}
{{- with .Params.videos }}
  {{- range . }}
    <meta property="og:video" content="{{ . | absURL }}">
  {{ end }}

{{ end }}

{{- /* If it is part of a series, link to related articles */}}
{{- $permalink := .Permalink }}
{{- $siteSeries := .Site.Taxonomies.series }}
{{ with .Params.series }}
  {{- range $name := . }}
    {{- $series := index $siteSeries ($name | urlize) }}
    {{- range $page := first 6 $series.Pages }}
      {{- if ne $page.Permalink $permalink }}
        <meta property="og:see_also" content="{{ $page.Permalink }}">
      {{ end }}
    {{- end }}
  {{ end }}

{{ end }}

{{- /* Facebook Page Admin ID for Domain Insights */}}
{{- with site.Params.social.facebook_admin }}
  <meta property="fb:admins" content="{{ . }}">
{{ end }}
