{{ $humanDate := time.Format "January 2, 2006" . }}
{{ $machineDate := time.Format "2006-01-02T15:04:05-07:00" . }}
<time {{ printf "datetime=%q" $machineDate | safeHTMLAttr }}>
  {{ $humanDate }}
</time>
