{{- /*
Renders the child sections of the given top-level section, listing each child's
immediate descendants.

@param {string} section The top-level section to render.

@example {{% quick-reference section="/functions" %}}
*/ -}}
{{ $section := "" }}
{{ with .Get "section" }}
  {{ $section = . }}
{{ else }}
  {{ errorf "The %q shortcode requires a 'section' parameter. See %s" .Name .Position }}
{{ end }}

{{ with site.GetPage $section }}
  {{ range .Sections }}
## {{ .LinkTitle }}{{/* Do not indent. */}}
{{ .Description }}{{/* Do not indent. */}}
    {{ .Content }}
    {{ with .Pages }}
      {{ range . }}
{{/* Use page Path as the link destination for render hook to resolve correctly. */}}
[{{ .LinkTitle }}]({{ .Path }}){{/* Do not indent. */}}
: {{ .Description }}{{/* Do not indent. */}}
      {{ end }}
    {{ end }}
  {{ end }}
{{ else }}
  {{ errorf "The %q shortcodes was unable to find the %q section. See %s" .Name $section .Position }}
{{ end }}
