<div class="overflow-x-auto">
  <table
    {{- range $k, $v := .Attributes }}
      {{- if $v }}
        {{- printf " %s=%q" $k $v | safeHTMLAttr }}
      {{- end }}
    {{- end }}>
    <thead>
      {{- range .THead }}
        <tr>
          {{- range . }}
            <th {{- with .Alignment }} class="!text-{{ . }}"{{- end }}>
              {{- .Text -}}
            </th>
          {{- end }}
        </tr>
      {{- end }}
    </thead>
    <tbody>
      {{- range .TBody }}
        <tr>
          {{- range . }}
            <td {{- with .Alignment }} class="!text-{{ . }}"{{- end }}>
              {{- .Text -}}
            </td>
          {{- end }}
        </tr>
      {{- end }}
    </tbody>
  </table>
</div>
