{{- with .Params.functions_and_methods.signatures }}
  <p class="font-bold text-dark dark:text-light">Syntax</p>
  {{- range . }}
    {{- $signature := . }}
    {{- if $.Params.function.returnType }}
      {{- $signature = printf "%s ⟼ %s" . $.Params.function.returnType }}
    {{- end }}
    <div class="font-sm font-mono ml-3 sm:ml-6 mt-0 sm:mt-1">
      {{- $signature -}}
    </div>
  {{- end }}
{{- end -}}
