<span class="h-3"></span>
<table class="mt-auto">
  <tbody>
    <tr>
      <td class="align-top text-red-700">weight:</td>
      <td class="align-top pl-2 w-full">
        {{ .Weight }}
      </td>
    </tr>
    <tr>
      <td class="align-top text-red-700">keywords:</td>
      <td class="align-top pl-2">
        {{ delimit (or .Keywords "") ", " }}
      </td>
    </tr>
    <tr>
      <td class="align-top text-red-700">categories:</td>
      <td class="align-top pl-2">
        {{ delimit (or .Params.categories "") ", " }}
      </td>
    </tr>
  </tbody>
</table>
