// Code generated by "stringer -type ItemType"; DO NOT EDIT.

package pageparser

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[tError-0]
	_ = x[tEOF-1]
	_ = x[TypeLeadSummaryDivider-2]
	_ = x[TypeFrontMatterYAML-3]
	_ = x[TypeFrontMatterTOML-4]
	_ = x[TypeFrontMatterJSON-5]
	_ = x[TypeFrontMatterORG-6]
	_ = x[TypeIgnore-7]
	_ = x[tLeftDelimScNoMarkup-8]
	_ = x[tRightDelimScNoMarkup-9]
	_ = x[tLeftDelimScWithMarkup-10]
	_ = x[tRightDelimScWithMarkup-11]
	_ = x[tScClose-12]
	_ = x[tScName-13]
	_ = x[tScNameInline-14]
	_ = x[tScParam-15]
	_ = x[tScParamVal-16]
	_ = x[tIndentation-17]
	_ = x[tText-18]
	_ = x[tKeywordMarker-19]
}

const _ItemType_name = "tErrortEOFTypeLeadSummaryDividerTypeFrontMatterYAMLTypeFrontMatterTOMLTypeFrontMatterJSONTypeFrontMatterORGTypeIgnoretLeftDelimScNoMarkuptRightDelimScNoMarkuptLeftDelimScWithMarkuptRightDelimScWithMarkuptScClosetScNametScNameInlinetScParamtScParamValtIndentationtTexttKeywordMarker"

var _ItemType_index = [...]uint16{0, 6, 10, 32, 51, 70, 89, 107, 117, 137, 158, 180, 203, 211, 218, 231, 239, 250, 262, 267, 281}

func (i ItemType) String() string {
	if i < 0 || i >= ItemType(len(_ItemType_index)-1) {
		return "ItemType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _ItemType_name[_ItemType_index[i]:_ItemType_index[i+1]]
}
