// Code generated by "stringer -type=NodeTransformState"; DO NOT EDIT.

package doctree

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[NodeTransformStateNone-0]
	_ = x[NodeTransformStateUpdated-1]
	_ = x[NodeTransformStateReplaced-2]
	_ = x[NodeTransformStateDeleted-3]
	_ = x[NodeTransformStateSkip-4]
	_ = x[NodeTransformStateTerminate-5]
}

const _NodeTransformState_name = "NodeTransformStateNoneNodeTransformStateUpdatedNodeTransformStateReplacedNodeTransformStateDeletedNodeTransformStateSkipNodeTransformStateTerminate"

var _NodeTransformState_index = [...]uint8{0, 22, 47, 73, 98, 120, 147}

func (i NodeTransformState) String() string {
	if i < 0 || i >= NodeTransformState(len(_NodeTransformState_index)-1) {
		return "NodeTransformState(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _NodeTransformState_name[_NodeTransformState_index[i]:_NodeTransformState_index[i+1]]
}
