{{- $images := partial "_funcs/get-page-images" . }}
{{- with index $images 0 }}
  <meta name="twitter:card" content="summary_large_image">
  <meta name="twitter:image" content="{{ .Permalink }}">
{{- else }}
  <meta name="twitter:card" content="summary">
{{- end }}

{{- with or .Title site.Title site.Params.title | plainify }}
  <meta name="twitter:title" content="{{ . }}">
{{- end }}

{{- with or .Description .Summary site.Params.description | plainify | htmlUnescape }}
  <meta name="twitter:description" content="{{ trim . "\n\r\t " }}">
{{- end }}

{{- $twitterSite := "" }}
{{- with site.Params.social }}
  {{- if reflect.IsMap . }}
    {{- with .twitter }}
      {{- $content := . }}
      {{- if not (strings.HasPrefix . "@") }}
        {{- $content = printf "@%v" . }}
      {{- end }}
      <meta name="twitter:site" content="{{ $content }}">
    {{- end }}
  {{- end }}
{{- end }}
