// Copyright 2024 The Hugo Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package resources

import (
	"os"
	"testing"

	qt "github.com/frankban/quicktest"
)

func TestAtomicStaler(t *testing.T) {
	c := qt.New(t)

	type test struct {
		AtomicStaler
	}

	var v test

	c.Assert(v.StaleVersion(), qt.Equals, uint32(0))
	v.MarkStale()
	c.Assert(v.StaleVersion(), qt.Equals, uint32(1))
	v.MarkStale()
	c.Assert(v.StaleVersion(), qt.Equals, uint32(2))
}

func BenchmarkHashImage(b *testing.B) {
	f, err := os.Open("testdata/sunset.jpg")
	if err != nil {
		b.Fatal(err)
	}
	defer f.Close()

	for b.Loop() {
		_, _, err := hashImage(f)
		if err != nil {
			b.Fatal(err)
		}
		f.Seek(0, 0)
	}
}
